/*
--	Copyright (c) 2007 - 2021 by Victor Shulman.
--	Written by Victor Shulman.  Not derived from licensed software.
--
--	Permission is granted to anyone to use this software for any
--	purpose on any computer system, and to redistribute it in any way,
--	subject to the following restrictions:
--
--	1. The author is not responsible for the consequences of use of
--		this software, no matter how awful, even if they arise
--		from defects in it.
--
--	2. The origin of this software must not be misrepresented, either
--		by explicit claim or by omission.
--
--	3. This notice must not be removed or altered.
*/
package com.dbtimes.nfldw

// import scala.collection.mutable

import java.util.Date

import org.apache.spark.storage.StorageLevel
// import scala.collection.mutable
import java.sql.{Date => SqlDate}
import java.util.Calendar
import org.apache.spark.sql.SparkSession

import com.typesafe.config.{Config, ConfigFactory}
import scala.collection.JavaConverters._

import com.dbtimes.dw.etl.ModelObject

/**
 * The purpose of this object it to initialize shared variables and DataFrames,
 * register UDFs, etc.
 *
 */
object DwNFL {
  val teamRoleOffense = "Offense"
  val teamRoleDefense = "Defense"
  val teamRolePenalty = "Penalty"
  val teamRoleTimeout = "Timeout"

  private var stgSourcesMonikerDescriptionTransactionName: List[(String, String, String)] = List.empty
  private var isDebug: Boolean = false

  def main(args: Array[String]) {

    val configFileName = if (args.length == 1) args(0) else "dwEtlDefault.json"
    val appConfig = ConfigFactory.load(configFileName)

    val sparkSessionBuilder = DwNFL.configureSparkSession(appConfig)
    val spark = sparkSessionBuilder.getOrCreate()

    // Do initialization specific to a given DW
    DwNFL.init(appConfig)

    // ETL process to build or update dimensions and fact tables
    ModelObject.runEtl(appConfig)

    spark.stop()
  }

  def getIsDebug: Boolean = isDebug
  def getStgSourcesMonikerDescriptionTransactionName: List[(String, String, String)] = stgSourcesMonikerDescriptionTransactionName

  private def configureSparkSession(appConfig: Config): SparkSession.Builder = {
    var sparkSessionBuilder = SparkSession.builder()
    sparkSessionBuilder = if (appConfig.hasPath("sparkParams.sessionAppName")) sparkSessionBuilder.appName(appConfig.getString("sparkParams.sessionAppName")) else sparkSessionBuilder
    sparkSessionBuilder = if (appConfig.hasPath("sparkParams.sessionConfigMaster")) sparkSessionBuilder.config("spark.master", appConfig.getString("sparkParams.sessionConfigMaster")) else sparkSessionBuilder
    //  .config("spark.driver.host", "localhost")  // to get rid of this error when running not connected to the internet. ERROR SparkContext: Error initializing SparkContext. java.lang.AssertionError: assertion failed: Expected hostname (not IP) but got 2600:1001:b028:7e92:b449:4c2d:7be2:d6e8
    sparkSessionBuilder = if (appConfig.hasPath("sparkParams.sessionConfigDriverHost")) sparkSessionBuilder.config("spark.driver.host", appConfig.getString("sparkParams.sessionConfigDriverHost")) else sparkSessionBuilder
    sparkSessionBuilder = if (appConfig.hasPath("sparkParams.sessionConfigLocalDir")) sparkSessionBuilder.config("spark.local.dir", appConfig.getString("sparkParams.sessionConfigLocalDir")) else sparkSessionBuilder

    sparkSessionBuilder = sparkSessionBuilder.config("spark.driver.maxResultSize", "5g")

    sparkSessionBuilder
  }

  private def init(appConfig: Config): Unit = {
    val spark = SparkSession.builder().getOrCreate() // this gets previously created session
    spark.udf.register("udfSeasonYear", DwNFL.getSeasonYear)

    val stgSources = appConfig.getConfigList("dwEtl.stgSources").asScala.toList
    stgSourcesMonikerDescriptionTransactionName = stgSources map {
      case source: Config => {
        (source.getString("moniker"),
          if (source.hasPath("description")) source.getString("description") else "",
          if (source.hasPath("transactionName")) source.getString("transactionName") else "")
      }
    }

    isDebug = if (appConfig.hasPath("isDebug")) {
      appConfig.getBoolean("isDebug")
    } else {
      false // default - no debugging
    }
  }

  private val getSeasonYear = (effDate: SqlDate) => {
    val cal = Calendar.getInstance()
    cal.setTime(effDate)
    val year = cal.get(Calendar.YEAR)
    val month = cal.get(Calendar.MONTH)
    if (month >= 8) year else year - 1
  }

}
