/*
--	Copyright (c) 2007 - 2021 by Victor Shulman.
--	Written by Victor Shulman.  Not derived from licensed software.
--
--	Permission is granted to anyone to use this software for any
--	purpose on any computer system, and to redistribute it in any way,
--	subject to the following restrictions:
--
--	1. The author is not responsible for the consequences of use of
--		this software, no matter how awful, even if they arise
--		from defects in it.
--
--	2. The origin of this software must not be misrepresented, either
--		by explicit claim or by omission.
--
--	3. This notice must not be removed or altered.
*/

package com.dbtimes.dw.etl

import java.sql.Timestamp
import java.nio.file.{FileSystems, Files, Path => JavaPath}
import java.text.SimpleDateFormat
import java.util.Calendar

import EffectiveDateRule.EffectiveDateRule
import FactProcessMode.FactProcessMode
import com.dbtimes.dw.common._
import com.typesafe.config.Config
import org.apache.spark.sql.types.{StructField, StructType}

import scala.collection.JavaConverters._
import scala.collection.immutable.ListMap

final private[etl] class ConfigDwEtl(private[etl] val configDwEtl: Config) extends ConfigDwJobsCommon {

  private[etl] def validateStgSourcesConfiguration(): Unit = {

    val stgSources = configDwEtl.getConfigList("dwEtl.stgSources").asScala.toList

    // Check: make sure the sources have "moniker" attribute defined
    {
      val errorSources = for (stgSource <- stgSources
                              if (!stgSource.hasPath("moniker"))
                              ) yield stgSource
      if (!errorSources.isEmpty)
        throw new RuntimeException(s"""DW ETL Configuration ERROR: Some staging sources do not have "moniker" attribute defined""")
    }

    // Check: make sure all sources are files, i.e., "fileSource" attribute defined
    {
      val errorSources = for {stgSource <- stgSources
                              if (!stgSource.hasPath("fileSource"))
                              } yield stgSource
      if (!errorSources.isEmpty)
        throw new RuntimeException(s"""DW ETL Configuration ERROR: All staging sources must be files, i.e., have "fileSource" defined. The following - do not: ${getSourceMonikers(errorSources).mkString(", ")}. """)
    }

    // Check: make sure the file sources have "fileSource.path" attribute defined
    {
      val errorSources = for {stgSource <- stgSources
                              if (stgSource.hasPath("fileSource") && !stgSource.hasPath("fileSource.path"))
                              } yield stgSource
      if (!errorSources.isEmpty)
        throw new RuntimeException(s"""DW ETL Configuration ERROR: All staging file sources must have "path" defined. The following - do not: ${getSourceMonikers(errorSources).mkString(", ")}. """)
    }

    // Check: make sure the sources are parquet files
    {
      val errorSources = for {stgSource <- stgSources // braces make semicolon after generator optional
                              sourceMoniker = stgSource.getString("moniker")
                              if (!getIsFileSourceParquet(sourceMoniker))
                              } yield stgSource
      if (!errorSources.isEmpty)
        throw new RuntimeException(s"""DW ETL Configuration ERROR: All staging sources must be parquet files. The following - are not: ${getSourceMonikers(errorSources).mkString(", ")}. """)
    }
  }

  private[etl] def validateModelConfiguration(): Unit = {


    // Dimension checks
    // -----------------------------------------------------------------

    // Check: make sure dimension names match pattern
    {
      val pattern = configDwEtl.getString("dwEtl.dimAuthority.dimensionNamePattern")
      val dimNamePatternRegEx = pattern.r
      val errorDims = for {dimName <- getDimNamesToLoad // check only dimensions that have "isLoad" flag set to true
                           firstMatchStrAsOption = dimNamePatternRegEx.findFirstIn(dimName)
                           if ( !firstMatchStrAsOption.isDefined || firstMatchStrAsOption.getOrElse("") != dimName )
                           } yield dimName
      if (!errorDims.isEmpty) {
        throw new RuntimeException(s"""DW ETL Configuration ERROR: the following dimension name(s) - ${errorDims.mkString(", ")} - do not match the dimension name pattern $pattern """)
      }
    }

    // Check: make sure each dimension has schema
    {
      val errorDims = for {dimName <- getDimNamesToLoad // check only dimensions that have "isLoad" flag set to true
                           dimCofig = getDimensionConfiguration(dimName)
                           if ( !dimCofig.hasPath("schema"))
                           } yield dimName
      if (!errorDims.isEmpty) {
        throw new RuntimeException(s"""DW ETL Configuration ERROR: the following dimension(s) - ${errorDims.mkString(", ")} - do not have "schema" defined """)
      }
    }

    // Check: each dimension must have dimension key
    {
      val errorDims = for {dimName <- getDimNamesToLoad // check only dimensions that have "isLoad" flag set to true
                           colsWithKey = super.getSchemaColNamesWithFlag(getDimensionConfiguration(dimName), "schema", "isKey", true, Some(false))
                           if (colsWithKey.length != 1)
                           } yield dimName
      if (!errorDims.isEmpty) {
        throw new RuntimeException(s"""DW ETL Configuration ERROR: the following dimension(s) - ${errorDims.mkString(", ")} - have none or more than one key column - a column with "IsKey" attribute. Each dimension must have one and only one key column """)
      }
    }

    // Check: Each dimension must have at least one natural key column
    {
      val errorDims = for (dimName <- getDimNamesToLoad; // check only dimensions that have "isLoad" flag set to true
                           naturalKeys = getDimNaturalKeys(dimName);
                           if (naturalKeys.length == 0 )
                           ) yield dimName
      if (!errorDims.isEmpty) {
        throw new RuntimeException(s"""DW ETL Configuration ERROR: Each dimension must have a natural key - at least one column with "isNaturalKey" flag set to true. The following dimension(s) - ${errorDims.mkString(", ")} - do not have natural key defined. """)
      }
    }

    // Check: If dimension key is a surrogate key, its type must be Long
    {
      val errorDims = for (dimName <- getDimNamesToLoad // check only dimensions that have "isLoad" flag set to true
                           if (getIsDimensionKeySurrogate(dimName) && getDimensionKeyColType(dimName) != "Long")
                           ) yield dimName
      if (!errorDims.isEmpty) {
        throw new RuntimeException(s"""DW ETL Configuration ERROR: the type of dimension key for dimensions ${errorDims.mkString(", ")} must be Long because the dimension key is a surrogate key """)
      }
    }

    // Check: Dimension key column must have unknown value
    // TODO

    // Check: Each natural key column must have unknown value
    {
      val errorDims = for (dimName <- getDimNamesToLoad; // check only dimensions that have "isLoad" flag set to true
                           naturalKeys = getDimNaturalKeys(dimName);
                           mapColNamesAndUnknownValues = getDimensionColNamesAndUnknownValues(dimName);
                           mapNaturalKeysAndUnknownValues = mapColNamesAndUnknownValues.filter((colAndValue) => getDimNaturalKeys(dimName).contains(colAndValue._1))
                           if (naturalKeys.length != mapNaturalKeysAndUnknownValues.size)
                           ) yield dimName
      if (!errorDims.isEmpty) {
        throw new RuntimeException(s"""DW ETL Configuration ERROR: All natural keys must have unknown value defined. Not all natural keys in dimension(s) ${errorDims.mkString(", ")} have unknown value defined """)
      }
    }


  }

  private[etl] def getDwClassName: Option[String] = {
    if (configDwEtl.hasPath("dwEtl.dwClassName")) Some(configDwEtl.getString("dwEtl.dwClassName")) else None
  }

  private[etl] def getDimAuthorityPackageName: String = configDwEtl.getString("dwEtl.dimAuthority.packageName")

  private[etl] def getDataMartPackageName: String = configDwEtl.getString("dwEtl.dataMart.packageName")

  private[etl] def getIsLoadDimensions: Boolean = {
    if ("LoadDims(Facts)?".r.findFirstMatchIn(configDwEtl.getString("dwEtl.jobType")).isDefined) true else false
  }

  private[etl] def getIsLoadFacts: Boolean = {
    if ("Load(Dims)?Facts".r.findFirstMatchIn(configDwEtl.getString("dwEtl.jobType")).isDefined) true else false
  }

  private def getDimensionConfiguration(dimName: String): Config = {
    val dimensions = configDwEtl.getConfigList("dwEtl.dimensions").asScala.toList
    val dimension = dimensions find {
      case dimension: Config => dimension.getString("name") == dimName
    }
    if (!dimension.isDefined)
      throw new RuntimeException("""ETL  ERROR: dimension """ + dimName + """ is not defined """)
    dimension.get
  }

  private[etl] def getDimNamesToLoad: List[String] = {
    val dimensions = configDwEtl.getConfigList("dwEtl.dimensions").asScala.toList
    val dimensionsToLoad = dimensions filter {
      case dimension: Config => {
        if ((dimension.hasPath("isLoad") && dimension.getBoolean("isLoad")) || !dimension.hasPath("isLoad")) // default is to load
          true
        else
          false
      }
    }
    dimensionsToLoad map { case dimension: Config => dimension.getString("name") } // return dimension name only
  }

  private[etl] def getIsDebugDwLib: Boolean = {
    if (configDwEtl.hasPath("dwEtl.isDebugDwLib")) {
      configDwEtl.getBoolean("dwEtl.isDebugDwLib")
    } else {
      false // default - no debugging
    }
  }

  private[etl] def getIsInitialLoad: Boolean = {
    configDwEtl.getBoolean("dwEtl.isInitialLoad")
  }

  private[etl] def getEtlLogFilePath: String = configDwEtl.getString("dwEtl.etlLogFilePath")

  private[etl] def getRerunEtlAfter: Option[String] = {
    if (configDwEtl.hasPath("dwEtl.rerunEtlAfter")) {
      Some(configDwEtl.getString("dwEtl.rerunEtlAfter"))
    } else {
      None
    }
  }

  ////////////////////////////////////////////////////////////////////////////
  ////////  Facts
  ////////////////////////////////////////////////////////////////////////////

  private[etl] def getFactNamesToLoad: List[String] = {
    val facts = configDwEtl.getConfigList("dwEtl.facts").asScala.toList
    val factsToLoad = facts filter {
      case fact: Config =>
        if ((fact.hasPath("isLoad") && fact.getBoolean("isLoad")) || !fact.hasPath("isLoad")) // default is to load
          true
        else
          false
    }
    factsToLoad map { case fact: Config => fact.getString("name") } // return fact name only
  }


  private[etl] def getStgSourceMonikersOfFact(factName: String): List[String] = {
    val factConfig = getFactConfiguration(factName)
    if (!factConfig.hasPath("factStgSources")) {
      List.empty[String]
    }
    else {
      val factStgSources = factConfig.getConfigList("factStgSources").asScala.toList
      factStgSources map {
        case stgSourceMoniker: Config => stgSourceMoniker.getString("moniker")
      }
    }
  }

  private def getFactConfiguration(factName: String): Config = {
    val facts = configDwEtl.getConfigList("dwEtl.facts").asScala.toList
    val fact = facts find {
      case fact: Config => fact.getString("name") == factName
    }
    if (!fact.isDefined) {
      throw new RuntimeException("""ETL  ERROR: fact """ + factName + """is not defined """)
    }
    fact.get
  }

  private[etl] def getFactSchema(factName: String): StructType = super.getSchema(getFactConfiguration(factName), "schema")

  private[etl] def getSchemaForMergeKeys(factName: String): StructType = {
    val mergeKeys = getMergeKeysCols(factName)

    val factSchema = getFactSchema(factName)

    val mergeKeysColumns = factSchema filter {
      case StructField(c, t, _, m) => if (mergeKeys.contains(c)) true else false
    }
    StructType(mergeKeysColumns)
  }


  private[etl] def getDimensionsForSettingForeignKeys(factName: String): List[String] = {

    val dimensionsForSettingForeignKeys = super.getSchemaDistinctPropertyValuesWithFlag(
      getFactConfiguration(factName), "schema", "underlyingDim", "isForSettingForeignKey", true, Some(false))

    dimensionsForSettingForeignKeys
  }

  private[etl] def getEffectiveDateColNameForSettingTypeTwoKeys(factName: String): Option[String] = {
    val effDateCols = super.getSchemaColNamesWithFlag(getFactConfiguration(factName), "schema", "isEffDateForTypeTwo", true, Some(false))

    if (effDateCols.isEmpty) None else Some(effDateCols.head)
  }

  /**
   * Iterate over all fact tables to be loaded and select all columns used to set key for this dimension.
   *
   * @param dimName
   * @return
   */
  private[etl] def getDimensionColsToSetKeyForAllFacts(dimName: String): List[String] = {

    val listOfColsFromThisDimensionToSetKeyForEachFact = for (
      factName <- getFactNamesToLoad;
      allDimensionsColsForSettingForeignKeys = super.getSchemaDistinctPropertyValuesWithOtherPropertyAndFlag(
        getFactConfiguration(factName), "schema", "underlyingDimCol",
        "underlyingDim", dimName, "isForSettingForeignKey", true, Some(false));
      dimensionCols = getDimCols(dimName);
      thisDimensionColsForSettingForeignKeys = allDimensionsColsForSettingForeignKeys filter {
        case dimensionColumn => dimensionCols.contains(dimensionColumn)
      }
    ) yield {
      thisDimensionColsForSettingForeignKeys
    }

    // combine all columns into one list. Because each fact table is likely to use the same columns
    // the columns in this list will be duplicated as many times as there are fact tables that use given dimension
    val listOfColsFromThisDimensionToSetKey = listOfColsFromThisDimensionToSetKeyForEachFact.reduceLeft((listOne, listTwo) => listOne ++ listTwo)

    listOfColsFromThisDimensionToSetKey.distinct
  }

  private[etl] def getPartitionByCols(factName: String): List[String] = {
    super.getSchemaColNamesWithFlag(getFactConfiguration(factName), "schema", "isForPartition", true, Some(false))
  }

  private[etl] def getMergeKeysCols(factName: String): List[String] = {
    super.getSchemaColNamesWithFlag(getFactConfiguration(factName), "schema", "isMergeKey", true, Some(false))
  }

  private[etl] def getForeignKeys(factName: String): List[String] = {
    super.getSchemaDistinctPropertyValues(getFactConfiguration(factName), "schema", "foreignKey");
  }

  private[etl] def getFactMeasures(factName: String): List[String] = {
    super.getSchemaColNamesWithFlag(getFactConfiguration(factName), "schema", "isMeasure", true, Some(false))
  }

  private[etl] def getDimensionForSettingForeignKey(factName: String, key: String): String = {
    val dimNameList = super.getSchemaDistinctPropertyValuesWithOtherPropertyAndFlag(
      getFactConfiguration(factName), "schema", "underlyingDim",
      "foreignKey", key, "isForSettingForeignKey", true, Some(false));

    // All names in the list should be the same
    dimNameList.head
  }

  private[etl] def getMapDimColsToFactCols(factName: String, key: String): Map[String, String] = {
    val dimCols = super.getSchemaDistinctPropertyValuesWithOtherPropertyAndFlag(
      getFactConfiguration(factName), "schema", "underlyingDimCol",
      "foreignKey", key, "isForSettingForeignKey", true, Some(false));
    val factCols = super.getSchemaDistinctPropertyValuesWithOtherPropertyAndFlag(
      getFactConfiguration(factName), "schema", "colName",
      "foreignKey", key, "isForSettingForeignKey", true, Some(false));

    val dimAndFactCols = dimCols zip factCols

    dimAndFactCols.toMap
  }

  private[etl] def getColsForSettingForeignKeys(factName: String): List[String] = {
    super.getSchemaColNamesWithFlag(
      getFactConfiguration(factName), "schema", "isForSettingForeignKey", true, Some(false))
  }

  private[etl] def getIsFactsDestinationFile: Boolean = if (configDwEtl.hasPath("dwEtl.dataMart.factsDestinationFile.parquet")) true else false

  private[etl] def getIsFactsDestinationFileParquet: Boolean = if (getIsFactsDestinationFile && configDwEtl.hasPath("dwEtl.dataMart.factsDestinationFile.parquet")) true else false

  private[etl] def getIsFactsDestinationParquet: Boolean = {
    if (configDwEtl.hasPath("dwEtl.dimAuthority.dimensionsDestinationFile.parquet")) true else false
  }

  private[etl] def getFactFilePath(factName: String, isForKeySetters: Boolean = false): String = {
    if (getIsFactsDestinationFile) {
      val factNamePatternRegEx = configDwEtl.getString("dwEtl.dataMart.factNamePattern").r
      val factFilePathWithPlaceholder = if (isForKeySetters) {
        val fileSeparator = FileSystems.getDefault().getSeparator()
        val loadTime = new SimpleDateFormat("yyyyMMdd_HH_mm_ss_SSS").format(Calendar.getInstance().getTimeInMillis())
        configDwEtl.getString("dwEtl.dataMart.factsDestinationFile.pathWithKeySetters") + fileSeparator + loadTime + "_" + ModelObject.jobId
      } else {
        configDwEtl.getString("dwEtl.dataMart.factsDestinationFile.path")
      }

      factNamePatternRegEx.replaceAllIn(factFilePathWithPlaceholder, factName)
    }
    else {
      throw new RuntimeException("""Etl Runner ERROR: Fact destination is not a file """)
    }

  }

  /**
   *
   * @return Map[String, (String, EffectiveDateRule, Option[String]) ] - a map of stg source moniker and a 3-tuple with
   *         1. the name of effective date column,
   *            2. the rule type of dates in that source
   *            3. Optional name of the Timestamp column
   */
  private[etl] def getFactProcessMode(factName: String): FactProcessMode = {
    val factConfig = getFactConfiguration(factName)

    if (factConfig.hasPath("processingMode")) {
      factConfig.getString("processingMode") match {
        case "REPLACE" => FactProcessMode.REPLACE
        case "REPLACE_PARTITION" => FactProcessMode.REPLACE_PARTITION
        case "MERGE" => FactProcessMode.MERGE
        case "MERGE_PARTITION" => FactProcessMode.MERGE_PARTITION
        case "ADD" => FactProcessMode.ADD
        case _ => if (getPartitionByCols(factName).isEmpty) FactProcessMode.REPLACE else FactProcessMode.REPLACE_PARTITION
      }
    } else {
      if (getPartitionByCols(factName).isEmpty) FactProcessMode.REPLACE else FactProcessMode.REPLACE_PARTITION // default
    } // _2
  }

  ////////////////////////////////////////////////////////////////////////////
  ////////  Dimensions
  ////////////////////////////////////////////////////////////////////////////

  private[etl] def getDimensionSchema(dimName: String): StructType = super.getSchema(getDimensionConfiguration(dimName), "schema")

  private[etl] def getDimensionSchemaWithoutKey(dimName: String): StructType = {
    val keyColName = getDimensionKeyCol(dimName)

    val dimSchema = getDimensionSchema(dimName)

    val columnsWithNoKey = dimSchema filter {
      case StructField(c, t, _, m) => if (c == keyColName) false else true
    }
    StructType(columnsWithNoKey)
  }

  private[etl] def getDimensionColNamesAndUnknownValues(dimName: String): ListMap[String, String] =
    super.getSchemaColNamesAndPropertyValues(getDimensionConfiguration(dimName), "schema", "unknownValue")

  private[etl] def getDimensionSurrogateKeyForUnknownRow(dimName: String, sep: String = "~"): String = {
    val mapColNamesAndUnknownValues = getDimensionColNamesAndUnknownValues(dimName)
    val mapNaturalKeysAndUnknownValues = mapColNamesAndUnknownValues.filter((colAndValue) => getDimNaturalKeys(dimName).contains(colAndValue._1))
    val surrogateKeyForUnknownRow = mapNaturalKeysAndUnknownValues.values.mkString(sep)
    surrogateKeyForUnknownRow
  }

  private[etl] def getIsDimDestinationFile: Boolean = if (configDwEtl.hasPath("dwEtl.dimAuthority.dimensionsDestinationFile.parquet")) true else false

  private[etl] def getIsDimDestinationFileParquet: Boolean = if (getIsDimDestinationFile && configDwEtl.hasPath("dwEtl.dimAuthority.dimensionsDestinationFile.parquet")) true else false

  private[etl] def getIsSavePreviousDimVersionOfDestinationFile: Boolean = if (!getIsDimDestinationFile || !configDwEtl.hasPath("dwEtl.dimAuthority.dimensionsDestinationFile.previousCopyPath")) false else true

  private[etl] def getDimensionFilePath(dimName: String): String = {
    if (getIsDimDestinationFile) {
      val dimNamePatternRegEx = configDwEtl.getString("dwEtl.dimAuthority.dimensionNamePattern").r
      val dimFilePathWithPlaceholder = configDwEtl.getString("dwEtl.dimAuthority.dimensionsDestinationFile.path")

      dimNamePatternRegEx.replaceAllIn(dimFilePathWithPlaceholder, dimName)
    }
    else {
      throw new RuntimeException("""Etl Runner ERROR: Dimension destination is not a file """)
    }

  }

  private[etl] def getDimensionPreviousVersionFilePath(dimName: String): String = {
    if (getIsDimDestinationFile) {
      val dimNamePatternRegEx = configDwEtl.getString("dwEtl.dimAuthority.dimensionNamePattern").r
      val dimFilePathWithPlaceholder = configDwEtl.getString("dwEtl.dimAuthority.dimensionsDestinationFile.previousCopyPath")

      dimNamePatternRegEx.replaceAllIn(dimFilePathWithPlaceholder, dimName)
    }
    else {
      throw new RuntimeException("""Etl Runner ERROR: Dimension destination is not a file """)
    }
  }

  /**
   *
   * @param dimName
   * @return the name of the dimension key. The can be only one dimension key column, so return the first one in the list
   */
  private[etl] def getDimensionKeyCol(dimName: String): String =
    super.getSchemaColNamesWithFlag(getDimensionConfiguration(dimName), "schema", "isKey", true, Some(false))(0)

  private[etl] def getDimensionKeyColIndex(dimName: String): Int =
    super.getSchemaColNames(getDimensionConfiguration(dimName), "schema").indexOf(getDimensionKeyCol(dimName))

  private[etl] def getIsDimensionKeySurrogate(dimName: String): Boolean = {
    if (super.getSchemaColNamesWithFlag(getDimensionConfiguration(dimName), "schema", "isSurrogateKey", true, None).isEmpty)
      false
    else
      true
  }

  private[etl] def getDimensionKeyColType(dimName: String): String = {
    val columnNameType = super.getSchemaColNamesAndPropertyValues(getDimensionConfiguration(dimName), "schema", "colType")
    columnNameType(getDimensionKeyCol(dimName))
  }

  private[etl] def getDimNaturalKeys(dimName: String): List[String] =
    super.getSchemaColNamesWithFlag(getDimensionConfiguration(dimName), "schema", "isNaturalKey", true, Some(false))

  private[etl] def getDimCols(dimName: String): List[String] = getDimensionSchema(dimName).fieldNames.toList

  private[etl] def getDimTypeTwoCols(dimName: String): List[String] =
    super.getSchemaColNamesWithFlag(getDimensionConfiguration(dimName), "schema", "isTypeTwo", true, Some(false))

  private[etl] def getDimTypeOneCols(dimName: String): List[String] = {
    super.getSchemaColNames(getDimensionConfiguration(dimName), "schema")
      .diff(List(getDimensionKeyCol(dimName)))
      .diff(getDimNaturalKeys(dimName))
      .diff(getDimTypeTwoCols(dimName))
  }

  private[etl] def getIsEffectiveDateRuleDistinct: Boolean = {
    if (configDwEtl.hasPath("dwEtl.dimAuthority.effectiveDateDays") &&
      configDwEtl.getString("dwEtl.dimAuthority.effectiveDateDays") == "DISTINCT") true else false
  }

  private[etl] def getIsEffectiveDateRuleWeekdays: Boolean = {
    if (configDwEtl.hasPath("dwEtl.dimAuthority.effectiveDateDays") &&
      configDwEtl.getString("dwEtl.dimAuthority.effectiveDateDays") == "WEEKDAYS") true else false
  }

  private[etl] def getIsEffectiveDateRuleAll: Boolean = {
    if ((configDwEtl.hasPath("dwEtl.dimAuthority.effectiveDateDays") &&
      configDwEtl.getString("dwEtl.dimAuthority.effectiveDateDays") == "ALL")
      || !getIsEffectiveDateRuleDistinct || !getIsEffectiveDateRuleWeekdays) true else false
  }

  private[etl] def getIsDimensionsDestinationParquet: Boolean = {
    if (configDwEtl.hasPath("dwEtl.dimAuthority.dimensionsDestinationFile.parquet")) true else false
  }

  private[etl] def getStgSourceMonikersOfDim(dimName: String): List[String] = {
    val dimConfig = getDimensionConfiguration(dimName)
    if (!dimConfig.hasPath("dimensionStgSources")) {
      List.empty[String]
    }
    else {
      val dimStgSources = dimConfig.getConfigList("dimensionStgSources").asScala.toList
      dimStgSources map {
        case stgSourceMoniker: Config => {
          stgSourceMoniker.getString("moniker")
        }
      }
    }
  }

  private[etl] def getDimNamesWhereStgSourceIsUsed(sourceMoniker: String): List[String] = {
    for (dimName <- getDimNamesToLoad if getStgSourceMonikersOfDim(dimName).contains(sourceMoniker)) yield dimName
  }

  ////////////////////////////////////////////////////////////////////////////
  ////////  Dims and Facts
  ////////////////////////////////////////////////////////////////////////////
  private[etl] def getStgSourceMonikersOfModelObject(modelObjectName: String): List[String] = {
    // At most one of these two methods will return a list. The final list may be empty if the object does not have any sources
    val dimNamePatternRegEx = configDwEtl.getString("dwEtl.dimAuthority.dimensionNamePattern").r
    val factNamePatternRegEx = configDwEtl.getString("dwEtl.dataMart.factNamePattern").r

    if (dimNamePatternRegEx.findFirstMatchIn(modelObjectName).isDefined) {
      getStgSourceMonikersOfDim(modelObjectName)
    }
    else if (factNamePatternRegEx.findFirstMatchIn(modelObjectName).isDefined) {
      getStgSourceMonikersOfFact(modelObjectName)
    }
    else {
      throw new RuntimeException("""ETL  ERROR: model object """ + modelObjectName + """does not match a name of a dimension or a fact """)
    }
  }

  ////////////////////////////////////////////////////////////////////////////
  ////////  Stg Sources
  ////////////////////////////////////////////////////////////////////////////
  private[etl] def getSourceMonikers: List[String] = {
    val stgSources = configDwEtl.getConfigList("dwEtl.stgSources").asScala.toList
    stgSources map {
      case source: Config => source.getString("moniker")
    }
  }

  private[etl] def getSourceMonikers( stgSources: List[Config] ): List[String] = {
    stgSources map {
      case source: Config => source.getString("moniker")
    }
  }

  /**
   *
   * @return Map[String, (String, EffectiveDateRule, Option[String]) ] - a map of stg source moniker and a 3-tuple with
   *         1. the name of effective date column,
   *         2. the rule type of dates in that source
   *         3. Optional name of the Timestamp column
   */
  private[etl] def getStgSourceMonikersForDeterminingEffectiveDates(modelObjectName: String): Map[String, (String, EffectiveDateRule, Option[String])] = {
    val stgSources = configDwEtl.getConfigList("dwEtl.stgSources").asScala.toList
    val stgSourceDefaultForEffectiveDate = stgSources filter {
      case source: Config => if (source.hasPath("effectiveDateColumn") && source.hasPath("isDefaultForEffectiveDate") && source.getBoolean("isDefaultForEffectiveDate")) true else false
    }

    val stgSourcesWithEffectiveDateForModelObject = stgSources filter {
      case source: Config => if (source.hasPath("effectiveDateColumn")
        && getStgSourceMonikersOfModelObject(modelObjectName).contains(source.getString("moniker"))) true else false
    }

    val stgSourcesWithEffectiveDate = if (stgSourcesWithEffectiveDateForModelObject.isEmpty) stgSourceDefaultForEffectiveDate else stgSourcesWithEffectiveDateForModelObject


    val stgSourceAndEffectiveDateProperties = stgSourcesWithEffectiveDate map {
      case source: Config =>
        source.getString("moniker") -> (
          source.getString("effectiveDateColumn"), // _1
          if (source.hasPath("effectiveDateRule")) {
            source.getString("effectiveDateRule") match {
              case "DISTINCT" => EffectiveDateRule.DISTINCT_DATES
              case "WEEKDAYS" => EffectiveDateRule.WEEKDAYS
              case _ => EffectiveDateRule.ALL_DATES
            }
          } else {
            EffectiveDateRule.ALL_DATES // default
          }, // _2
          if (source.hasPath("timestampColumn")) {
            Some(source.getString("timestampColumn"))
          } else {
            None // default
          } // _3
        )
    }

    stgSourceAndEffectiveDateProperties.toMap
  }

  private[etl] def getStgSourceTimestampColumn(sourceMoniker: String): Option[String] = {
    if (getSourceConfiguration(sourceMoniker).hasPath("timestampColumn")) {
      Some(getSourceConfiguration(sourceMoniker).getString("timestampColumn"))
    } else {
      None
    }
  }

  private def getSourceConfiguration(sourceMoniker: String): Config = {
    val stgSources = configDwEtl.getConfigList("dwEtl.stgSources").asScala.toList
    val stgSource = stgSources find {
      case source: Config => {
        source.getString("moniker") == sourceMoniker
      }
    }
    if (!stgSource.isDefined)
      throw new RuntimeException("""ETL  ERROR: etl source """ + sourceMoniker + """is not defined """)
    stgSource.get
  }

  private[etl] def getIsFileSourceParquet(sourceMoniker: String): Boolean = {
    if (getSourceConfiguration(sourceMoniker).hasPath("fileSource.parquet"))
      true
    else
      false
  }

  private[etl] def getSourceFilePath(sourceMoniker: String): String = {
    getSourceConfiguration(sourceMoniker).getString("fileSource.path")
  }

}
