/*
--	Copyright (c) 2007 - 2021 by Victor Shulman.
--	Written by Victor Shulman.  Not derived from licensed software.
--
--	Permission is granted to anyone to use this software for any
--	purpose on any computer system, and to redistribute it in any way,
--	subject to the following restrictions:
--
--	1. The author is not responsible for the consequences of use of
--		this software, no matter how awful, even if they arise
--		from defects in it.
--
--	2. The origin of this software must not be misrepresented, either
--		by explicit claim or by omission.
--
--	3. This notice must not be removed or altered.
*/
package com.dbtimes.dw.sourceloader

import com.dbtimes.dw.common._
import org.apache.spark.sql.SparkSession
import java.util.Properties
import org.apache.spark.sql.functions._
import org.apache.spark.sql.types._

import LogFile.{logger => loaderLog}

private[sourceloader] class ProcessorOfSqlServerSource(private val action: LoadAction) extends ActionProcessor
  with SourceDataMerger {

  private[sourceloader] def process(): Unit = {
    val spark = SparkSession.builder().getOrCreate() // this gets previously created session

    Class.forName("com.microsoft.sqlserver.jdbc.SQLServerDriver")

    // Create a Properties() object to hold the parameters.
    val connectionProperties = new Properties()
    val driverClass = "com.microsoft.sqlserver.jdbc.SQLServerDriver"
    connectionProperties.setProperty("Driver", driverClass)
    if (action.getUser.isDefined && action.getPassword.isDefined) {
      connectionProperties.put("user", action.getUser.get)
      connectionProperties.put("password", action.getPassword.get)
    }
    val jdbcUrl = action.getJdbcUrl.get


    // Get data for incremental load if incremental load attributes are available
    val isDoIncrementalLoad = action.isIncrementalLoad

    val sourceTableChangesSqlAsOption = if (isDoIncrementalLoad) {

      val columnNameAndPlaceholderPairs = action.getIncrementalLoadColumnNameAndPlaceholderPairs

      // 1. first get MAX values for each column name from the existing file
      // The value is string represenation of the column's values, i.e. it is cast to String
      val placeholderAndMaxValues = for ((colName, placeholder) <- columnNameAndPlaceholderPairs;
                                         dfMaxNativeColumnType = spark.read.format("org.apache.spark.sql.execution.datasources.parquet.ParquetFileFormat")
                                           .load(action.getSdaDestinationFilePath)
                                           .agg(max(colName));
                                         dfMaxAsString = dfMaxNativeColumnType.withColumn(dfMaxNativeColumnType.schema.fieldNames(0), col(dfMaxNativeColumnType.schema.fieldNames(0)).cast(StringType));
                                         maxValueAsString = dfMaxAsString.head.getString(0)
                                         ) yield (placeholder, maxValueAsString)


      // 2. replace all placeholders with max values
      val sourceTableChangesSql = placeholderAndMaxValues.foldLeft(action.getIncrementalLoadSourceTableChangesSql) {
        case (currentSql, (placeholder, maxValue)) => currentSql.replaceAll(placeholder, maxValue)
      }

      loaderLog.info("-- Executing sql to get tabl changes: " + sourceTableChangesSql)

      Some(sourceTableChangesSql)
    } else None

    val minValue = if (action.getPartitionColumn.isDefined) { // load using partitions
      spark.read.jdbc(
        jdbcUrl,
        "( SELECT CAST( ISNULL( MIN( [<@PartitionColumn>] ), 0 ) AS BIGINT) AS Result FROM <@SourceTable> ) AS result"
          .replace("<@PartitionColumn>", action.getPartitionColumn.get)
          .replace("<@SourceTable>", if (isDoIncrementalLoad) sourceTableChangesSqlAsOption.get else action.getSourceTable.get),
        connectionProperties).head.getLong(0)
    } else 0L

    val maxValue = if (action.getPartitionColumn.isDefined) {
      spark.read.jdbc(
        jdbcUrl,
        "( SELECT CAST( ISNULL( MAX( [<@PartitionColumn>] ), 1 ) AS BIGINT)  AS Result FROM <@SourceTable> ) AS result"
          .replace("<@PartitionColumn>", action.getPartitionColumn.get)
          .replace("<@SourceTable>", if (isDoIncrementalLoad) sourceTableChangesSqlAsOption.get else action.getSourceTable.get),
        connectionProperties).head.getLong(0)
    } else 0L

    val dfNewData = if (action.getPartitionColumn.isDefined) { // load using partitions
      spark.read.jdbc(
        jdbcUrl,
        if (isDoIncrementalLoad) sourceTableChangesSqlAsOption.get else action.getSourceTable.get,
        action.getPartitionColumn.get,
        minValue,
        maxValue,
        action.getNumberOfPartitions,
        connectionProperties)
    }
    else {
      spark.read.jdbc(jdbcUrl, if (isDoIncrementalLoad) sourceTableChangesSqlAsOption.get else action.getSourceTable.get, connectionProperties)
    }

    //    loaderLog.info("** Processing file " + sourceFilePath )
    val effectiveDate = action.getEffectiveDate(dfNewData)

    // Read merge keys. Need this to handle deleted records
    if (isDoIncrementalLoad) {
      val dfAllMergeKeys = if (action.getPartitionColumn.isDefined && action.getSdaMergeKeysList.contains(action.getPartitionColumn.get)) {
        spark.read.jdbc(
          jdbcUrl,
          action.getIncrementalLoadSourceTableUniqueKeysSql.getOrElse( buildQueryToGetAllUniqueKeys(action)),
          action.getPartitionColumn.get,
          minValue,
          maxValue,
          action.getNumberOfPartitions,
          connectionProperties)
      }
      else {
        spark.read.jdbc(jdbcUrl, action.getIncrementalLoadSourceTableUniqueKeysSql.getOrElse( buildQueryToGetAllUniqueKeys(action)), connectionProperties)
      }

      mergeSourceDataChanges(action, dfNewData, isNewDataPreparedForMerge = false, dfAllMergeKeys, effectiveDate, loaderLog)
    } else {
      mergeSourceData(action, dfNewData, effectiveDate, loaderLog)
    }
  }

  private[sourceloader] def buildQueryToGetAllUniqueKeys(action: LoadAction): String = {

    // The query will look like this
    // ( SELECT list-of-unique-keys FROM [source-table] ) AS table-alias
    // this should work in any DBMS

    val query = s"""( SELECT ${action.getSdaPrimaryKeysList.mkString("[", "], [", "]")}  FROM ${action.getSourceTable.get} ) AS tbl${java.util.UUID.randomUUID.toString.replace("-", "")} """
    query
  }
}
